@@ -9,7 +9,9 @@
 #include <nxt_runtime.h>
 #include <nxt_main_process.h>
 #include <nxt_conf.h>
+#include <nxt_status.h>
 #include <nxt_cert.h>
+#include <nxt_script.h>
 
 
 typedef struct {
@@ -85,6 +87,12 @@ static void nxt_controller_process_request(nxt_task_t *task,
 static void nxt_controller_process_config(nxt_task_t *task,
     nxt_controller_request_t *req, nxt_str_t *path);
 static nxt_bool_t nxt_controller_check_postpone_request(nxt_task_t *task);
+static void nxt_controller_process_status(nxt_task_t *task,
+    nxt_controller_request_t *req);
+static void nxt_controller_status_handler(nxt_task_t *task,
+    nxt_port_recv_msg_t *msg, void *data);
+static void nxt_controller_status_response(nxt_task_t *task,
+    nxt_controller_request_t *req, nxt_str_t *path);
 #if (NXT_TLS)
 static void nxt_controller_process_cert(nxt_task_t *task,
     nxt_controller_request_t *req, nxt_str_t *path);
@@ -94,6 +102,15 @@ static nxt_bool_t nxt_controller_cert_in_use(nxt_str_t *name);
 static void nxt_controller_cert_cleanup(nxt_task_t *task, void *obj,
     void *data);
 #endif
+#if (NXT_HAVE_NJS)
+static void nxt_controller_process_script(nxt_task_t *task,
+    nxt_controller_request_t *req, nxt_str_t *path);
+static void nxt_controller_process_script_save(nxt_task_t *task,
+    nxt_port_recv_msg_t *msg, void *data);
+static nxt_bool_t nxt_controller_script_in_use(nxt_str_t *name);
+static void nxt_controller_script_cleanup(nxt_task_t *task, void *obj,
+    void *data);
+#endif
 static void nxt_controller_process_control(nxt_task_t *task,
     nxt_controller_request_t *req, nxt_str_t *path);
 static void nxt_controller_app_restart_handler(nxt_task_t *task,
@@ -120,6 +137,7 @@ static nxt_uint_t              nxt_controller_router_ready;
 static nxt_controller_conf_t   nxt_controller_conf;
 static nxt_queue_t             nxt_controller_waiting_requests;
 static nxt_bool_t              nxt_controller_waiting_init_conf;
+static nxt_conf_value_t        *nxt_controller_status;
 
 
 static const nxt_event_conn_state_t  nxt_controller_conn_read_state;
@@ -205,6 +223,13 @@ nxt_controller_prefork(nxt_task_t *task, nxt_process_t *process, nxt_mp_t *mp)
     nxt_mp_cleanup(mp, nxt_controller_cert_cleanup, task, ctrl_init.certs, rt);
 #endif
 
+#if (NXT_HAVE_NJS)
+    ctrl_init.scripts = nxt_script_store_load(task, mp);
+
+    nxt_mp_cleanup(mp, nxt_controller_script_cleanup, task, ctrl_init.scripts,
+                   rt);
+#endif
+
     process->data.controller = ctrl_init;
 
     return NXT_OK;
@@ -313,6 +338,13 @@ nxt_controller_start(nxt_task_t *task, nxt_process_data_t *data)
     }
 #endif
 
+#if (NXT_HAVE_NJS)
+    if (init->scripts != NULL) {
+        nxt_script_info_init(task, init->scripts);
+        nxt_script_store_release(init->scripts);
+    }
+#endif
+
     json = &init->conf;
 
     if (json->start == NULL) {
@@ -478,8 +510,9 @@ nxt_controller_conf_default(void)
     nxt_mp_t          *mp;
     nxt_conf_value_t  *conf;
 
-    static const nxt_str_t json
-        = nxt_string("{ \"listeners\": {}, \"applications\": {} }");
+    static const nxt_str_t json = nxt_string(
+        "{ \"listeners\": {}, \"routes\": [], \"applications\": {} }"
+    );
 
     mp = nxt_mp_create(1024, 128, 256, 32);
 
@@ -658,6 +691,14 @@ nxt_runtime_controller_socket(nxt_task_t *task, nxt_runtime_t *rt)
 #endif
     ls->handler = nxt_controller_conn_init;
 
+#if (NXT_HAVE_UNIX_DOMAIN)
+    if (ls->sockaddr->u.sockaddr.sa_family == AF_UNIX) {
+        const char *path = ls->sockaddr->u.sockaddr_un.sun_path;
+
+        nxt_fs_mkdir_p_dirname((const u_char *) path, 0755);
+    }
+#endif
+
     if (nxt_listen_socket_create(task, rt->mem_pool, ls) != NXT_OK) {
         return NXT_ERROR;
     }
@@ -725,7 +766,7 @@ static const nxt_event_conn_state_t  nxt_controller_conn_read_state
 
     .timer_handler = nxt_controller_conn_read_timeout,
     .timer_value = nxt_controller_conn_timeout_value,
-    .timer_data = 60 * 1000,
+    .timer_data = 300 * 1000,
 };
 
 
@@ -1031,10 +1072,21 @@ nxt_controller_process_request(nxt_task_t *task, nxt_controller_request_t *req)
     nxt_controller_response_t  resp;
 #if (NXT_TLS)
     nxt_conf_value_t           *certs;
+#endif
+#if (NXT_HAVE_NJS)
+    nxt_conf_value_t           *scripts;
+#endif
 
+#if (NXT_TLS)
     static nxt_str_t certificates = nxt_string("certificates");
 #endif
+
+#if (NXT_HAVE_NJS)
+    static nxt_str_t scripts_str = nxt_string("js_modules");
+#endif
+
     static nxt_str_t config = nxt_string("config");
+    static nxt_str_t status = nxt_string("status");
 
     c = req->conn;
     path = req->parser.path;
@@ -1058,6 +1110,32 @@ nxt_controller_process_request(nxt_task_t *task, nxt_controller_request_t *req)
         return;
     }
 
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    if (nxt_str_start(&path, "/status", 7)
+        && (path.length == 7 || path.start[7] == '/'))
+    {
+        if (!nxt_str_eq(&req->parser.method, "GET", 3)) {
+            goto invalid_method;
+        }
+
+        if (nxt_controller_status == NULL) {
+            nxt_controller_process_status(task, req);
+            return;
+        }
+
+        if (path.length == 7) {
+            path.length = 1;
+
+        } else {
+            path.length -= 7;
+            path.start += 7;
+        }
+
+        nxt_controller_status_response(task, req, &path);
+        return;
+    }
+
 #if (NXT_TLS)
 
     if (nxt_str_start(&path, "/certificates", 13)
@@ -1075,6 +1153,25 @@ nxt_controller_process_request(nxt_task_t *task, nxt_controller_request_t *req)
         return;
     }
 
+#endif
+
+#if (NXT_HAVE_NJS)
+
+    if (nxt_str_start(&path, "/js_modules", 11)
+        && (path.length == 11 || path.start[11] == '/'))
+    {
+        if (path.length == 11) {
+            path.length = 1;
+
+        } else {
+            path.length -= 11;
+            path.start += 11;
+        }
+
+        nxt_controller_process_script(task, req, &path);
+        return;
+    }
+
 #endif
 
     if (nxt_str_start(&path, "/control/", 9)) {
@@ -1085,18 +1182,24 @@ nxt_controller_process_request(nxt_task_t *task, nxt_controller_request_t *req)
         return;
     }
 
-    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
-
     if (path.length == 1 && path.start[0] == '/') {
 
         if (!nxt_str_eq(&req->parser.method, "GET", 3)) {
             goto invalid_method;
         }
 
-        count = 1;
+        if (nxt_controller_status == NULL) {
+            nxt_controller_process_status(task, req);
+            return;
+        }
+
+        count = 2;
 #if (NXT_TLS)
         count++;
 #endif
+#if (NXT_HAVE_NJS)
+        count++;
+#endif
 
         value = nxt_conf_create_object(c->mem_pool, count);
         if (nxt_slow_path(value == NULL)) {
@@ -1114,7 +1217,17 @@ nxt_controller_process_request(nxt_task_t *task, nxt_controller_request_t *req)
         nxt_conf_set_member(value, &certificates, certs, i++);
 #endif
 
-        nxt_conf_set_member(value, &config, nxt_controller_conf.root, i);
+#if (NXT_HAVE_NJS)
+        scripts = nxt_script_info_get_all(c->mem_pool);
+        if (nxt_slow_path(scripts == NULL)) {
+            goto alloc_fail;
+        }
+
+        nxt_conf_set_member(value, &scripts_str, scripts, i++);
+#endif
+
+        nxt_conf_set_member(value, &config, nxt_controller_conf.root, i++);
+        nxt_conf_set_member(value, &status, nxt_controller_status, i);
 
         resp.status = 200;
         resp.conf = value;
@@ -1216,7 +1329,7 @@ nxt_controller_process_config(nxt_task_t *task, nxt_controller_request_t *req,
 
         /* Skip UTF-8 BOM. */
         if (nxt_buf_mem_used_size(mbuf) >= 3
-            && nxt_memcmp(mbuf->pos, "\xEF\xBB\xBF", 3) == 0)
+            && memcmp(mbuf->pos, "\xEF\xBB\xBF", 3) == 0)
         {
             mbuf->pos += 3;
         }
@@ -1451,6 +1564,128 @@ nxt_controller_check_postpone_request(nxt_task_t *task)
 }
 
 
+static void
+nxt_controller_process_status(nxt_task_t *task, nxt_controller_request_t *req)
+{
+    uint32_t                   stream;
+    nxt_int_t                  rc;
+    nxt_port_t                 *router_port, *controller_port;
+    nxt_runtime_t              *rt;
+    nxt_controller_response_t  resp;
+
+    if (nxt_controller_check_postpone_request(task)) {
+        nxt_queue_insert_tail(&nxt_controller_waiting_requests, &req->link);
+        return;
+    }
+
+    rt = task->thread->runtime;
+
+    router_port = rt->port_by_type[NXT_PROCESS_ROUTER];
+
+    nxt_assert(router_port != NULL);
+    nxt_assert(nxt_controller_router_ready);
+
+    controller_port = rt->port_by_type[NXT_PROCESS_CONTROLLER];
+
+    stream = nxt_port_rpc_register_handler(task, controller_port,
+                                           nxt_controller_status_handler,
+                                           nxt_controller_status_handler,
+                                           router_port->pid, req);
+    if (nxt_slow_path(stream == 0)) {
+        goto fail;
+    }
+
+    rc = nxt_port_socket_write(task, router_port, NXT_PORT_MSG_STATUS,
+                               -1, stream, controller_port->id, NULL);
+
+    if (nxt_slow_path(rc != NXT_OK)) {
+        nxt_port_rpc_cancel(task, controller_port, stream);
+
+        goto fail;
+    }
+
+    nxt_queue_insert_head(&nxt_controller_waiting_requests, &req->link);
+    return;
+
+fail:
+
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    resp.status = 500;
+    resp.title = (u_char *) "Failed to get status.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+}
+
+
+static void
+nxt_controller_status_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg,
+    void *data)
+{
+    nxt_conf_value_t           *status;
+    nxt_controller_request_t   *req;
+    nxt_controller_response_t  resp;
+
+    nxt_debug(task, "controller status handler");
+
+    req = data;
+
+    if (msg->port_msg.type == NXT_PORT_MSG_RPC_READY) {
+        status = nxt_status_get((nxt_status_report_t *) msg->buf->mem.pos,
+                                req->conn->mem_pool);
+    } else {
+        status = NULL;
+    }
+
+    if (status == NULL) {
+        nxt_queue_remove(&req->link);
+
+        nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+        resp.status = 500;
+        resp.title = (u_char *) "Failed to get status.";
+        resp.offset = -1;
+
+        nxt_controller_response(task, req, &resp);
+    }
+
+    nxt_controller_status = status;
+
+    nxt_controller_flush_requests(task);
+
+    nxt_controller_status = NULL;
+}
+
+
+static void
+nxt_controller_status_response(nxt_task_t *task, nxt_controller_request_t *req,
+    nxt_str_t *path)
+{
+    nxt_conf_value_t           *status;
+    nxt_controller_response_t  resp;
+
+    status = nxt_conf_get_path(nxt_controller_status, path);
+
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    if (status == NULL) {
+        resp.status = 404;
+        resp.title = (u_char *) "Invalid path.";
+        resp.offset = -1;
+
+        nxt_controller_response(task, req, &resp);
+        return;
+    }
+
+    resp.status = 200;
+    resp.conf = status;
+
+    nxt_controller_response(task, req, &resp);
+}
+
+
 #if (NXT_TLS)
 
 static void
@@ -1468,7 +1703,7 @@ nxt_controller_process_cert(nxt_task_t *task,
     name.length = path->length - 1;
     name.start = path->start + 1;
 
-    p = nxt_memchr(name.start, '/', name.length);
+    p = memchr(name.start, '/', name.length);
 
     if (p != NULL) {
         name.length = p - name.start;
@@ -1673,12 +1908,12 @@ nxt_controller_process_cert_save(nxt_task_t *task, nxt_port_recv_msg_t *msg,
 static nxt_bool_t
 nxt_controller_cert_in_use(nxt_str_t *name)
 {
-    uint32_t          next;
+    uint32_t          i, n, next;
     nxt_str_t         str;
-    nxt_conf_value_t  *listeners, *listener, *value;
+    nxt_conf_value_t  *listeners, *listener, *value, *element;
 
-    static nxt_str_t  listeners_path = nxt_string("/listeners");
-    static nxt_str_t  certificate_path = nxt_string("/tls/certificate");
+    static const nxt_str_t  listeners_path = nxt_string("/listeners");
+    static const nxt_str_t  certificate_path = nxt_string("/tls/certificate");
 
     listeners = nxt_conf_get_path(nxt_controller_conf.root, &listeners_path);
 
@@ -1696,7 +1931,294 @@ nxt_controller_cert_in_use(nxt_str_t *name)
                 continue;
             }
 
-            nxt_conf_get_string(value, &str);
+            if (nxt_conf_type(value) == NXT_CONF_ARRAY) {
+                n = nxt_conf_array_elements_count(value);
+
+                for (i = 0; i < n; i++) {
+                    element = nxt_conf_get_array_element(value, i);
+
+                    nxt_conf_get_string(element, &str);
+
+                    if (nxt_strstr_eq(&str, name)) {
+                        return 1;
+                    }
+                }
+
+            } else {
+                /* NXT_CONF_STRING */
+
+                nxt_conf_get_string(value, &str);
+
+                if (nxt_strstr_eq(&str, name)) {
+                    return 1;
+                }
+            }
+        }
+    }
+
+    return 0;
+}
+
+#endif
+
+
+#if (NXT_HAVE_NJS)
+
+static void
+nxt_controller_process_script(nxt_task_t *task,
+    nxt_controller_request_t *req, nxt_str_t *path)
+{
+    u_char                     *p;
+    nxt_int_t                  ret;
+    nxt_str_t                  name;
+    nxt_conn_t                 *c;
+    nxt_script_t               *script;
+    nxt_buf_mem_t              *bm;
+    nxt_conf_value_t           *value;
+    nxt_controller_response_t  resp;
+    u_char                     error[NXT_MAX_ERROR_STR];
+
+    name.length = path->length - 1;
+    name.start = path->start + 1;
+
+    p = memchr(name.start, '/', name.length);
+
+    if (p != NULL) {
+        name.length = p - name.start;
+
+        path->length -= p - path->start;
+        path->start = p;
+
+    } else {
+        path = NULL;
+    }
+
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    c = req->conn;
+
+    if (nxt_str_eq(&req->parser.method, "GET", 3)) {
+
+        if (name.length != 0) {
+            value = nxt_script_info_get(&name);
+            if (value == NULL) {
+                goto script_not_found;
+            }
+
+            if (path != NULL) {
+                value = nxt_conf_get_path(value, path);
+                if (value == NULL) {
+                    goto not_found;
+                }
+            }
+
+        } else {
+            value = nxt_script_info_get_all(c->mem_pool);
+            if (value == NULL) {
+                goto alloc_fail;
+            }
+        }
+
+        resp.status = 200;
+        resp.conf = value;
+
+        nxt_controller_response(task, req, &resp);
+        return;
+    }
+
+    if (name.length == 0 || path != NULL) {
+        goto invalid_name;
+    }
+
+    if (nxt_str_eq(&req->parser.method, "PUT", 3)) {
+        value = nxt_script_info_get(&name);
+        if (value != NULL) {
+            goto exists_script;
+        }
+
+        bm = &c->read->mem;
+
+        script = nxt_script_new(task, &name, bm->pos,
+                                nxt_buf_mem_used_size(bm), error);
+        if (script == NULL) {
+            goto invalid_script;
+        }
+
+        ret = nxt_script_info_save(&name, script);
+
+        nxt_script_destroy(script);
+
+        if (nxt_slow_path(ret != NXT_OK)) {
+            goto alloc_fail;
+        }
+
+        nxt_script_store_get(task, &name, c->mem_pool,
+                             nxt_controller_process_script_save, req);
+        return;
+    }
+
+    if (nxt_str_eq(&req->parser.method, "DELETE", 6)) {
+
+        if (nxt_controller_script_in_use(&name)) {
+            goto script_in_use;
+        }
+
+        if (nxt_script_info_delete(&name) != NXT_OK) {
+            goto script_not_found;
+        }
+
+        nxt_script_store_delete(task, &name, c->mem_pool);
+
+        resp.status = 200;
+        resp.title = (u_char *) "JS module deleted.";
+
+        nxt_controller_response(task, req, &resp);
+        return;
+    }
+
+    resp.status = 405;
+    resp.title = (u_char *) "Invalid method.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+invalid_name:
+
+    resp.status = 400;
+    resp.title = (u_char *) "Invalid JS module name.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+invalid_script:
+
+    resp.status = 400;
+    resp.title = (u_char *) "Invalid JS module.";
+    resp.offset = -1;
+
+    resp.detail.start = error;
+    resp.detail.length = nxt_strlen(error);
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+exists_script:
+
+    resp.status = 400;
+    resp.title = (u_char *) "JS module already exists.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+script_in_use:
+
+    resp.status = 400;
+    resp.title = (u_char *) "JS module is used in the configuration.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+script_not_found:
+
+    resp.status = 404;
+    resp.title = (u_char *) "JS module doesn't exist.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+not_found:
+
+    resp.status = 404;
+    resp.title = (u_char *) "Invalid path.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+    return;
+
+alloc_fail:
+
+    resp.status = 500;
+    resp.title = (u_char *) "Memory allocation failed.";
+    resp.offset = -1;
+
+    nxt_controller_response(task, req, &resp);
+}
+
+
+static void
+nxt_controller_process_script_save(nxt_task_t *task, nxt_port_recv_msg_t *msg,
+    void *data)
+{
+    nxt_conn_t                 *c;
+    nxt_buf_mem_t              *mbuf;
+    nxt_controller_request_t   *req;
+    nxt_controller_response_t  resp;
+
+    req = data;
+
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    if (msg == NULL || msg->port_msg.type == _NXT_PORT_MSG_RPC_ERROR) {
+        resp.status = 500;
+        resp.title = (u_char *) "Failed to store script.";
+
+        nxt_controller_response(task, req, &resp);
+        return;
+    }
+
+    c = req->conn;
+
+    mbuf = &c->read->mem;
+
+    nxt_fd_write(msg->fd[0], mbuf->pos, nxt_buf_mem_used_size(mbuf));
+
+    nxt_fd_close(msg->fd[0]);
+
+    nxt_memzero(&resp, sizeof(nxt_controller_response_t));
+
+    resp.status = 200;
+    resp.title = (u_char *) "JS module uploaded.";
+
+    nxt_controller_response(task, req, &resp);
+}
+
+
+static nxt_bool_t
+nxt_controller_script_in_use(nxt_str_t *name)
+{
+    uint32_t          i, n;
+    nxt_str_t         str;
+    nxt_conf_value_t  *js_module, *element;
+
+    static const nxt_str_t  js_module_path = nxt_string("/settings/js_module");
+
+    js_module = nxt_conf_get_path(nxt_controller_conf.root,
+                                    &js_module_path);
+
+    if (js_module != NULL) {
+
+        if (nxt_conf_type(js_module) == NXT_CONF_ARRAY) {
+            n = nxt_conf_array_elements_count(js_module);
+
+            for (i = 0; i < n; i++) {
+                element = nxt_conf_get_array_element(js_module, i);
+
+                nxt_conf_get_string(element, &str);
+
+                if (nxt_strstr_eq(&str, name)) {
+                    return 1;
+                }
+            }
+
+        } else {
+            /* NXT_CONF_STRING */
+
+            nxt_conf_get_string(js_module, &str);
 
             if (nxt_strstr_eq(&str, name)) {
                 return 1;
@@ -1707,6 +2229,24 @@ nxt_controller_cert_in_use(nxt_str_t *name)
     return 0;
 }
 
+
+static void
+nxt_controller_script_cleanup(nxt_task_t *task, void *obj, void *data)
+{
+    pid_t          main_pid;
+    nxt_array_t    *scripts;
+    nxt_runtime_t  *rt;
+
+    scripts = obj;
+    rt = data;
+
+    main_pid = rt->port_by_type[NXT_PROCESS_MAIN]->pid;
+
+    if (nxt_pid == main_pid && scripts != NULL) {
+        nxt_script_store_release(scripts);
+    }
+}
+
 #endif
 
 
@@ -1771,7 +2311,7 @@ nxt_controller_process_control(nxt_task_t *task,
     }
 
     if (!nxt_str_start(path, "applications/", 13)
-        || nxt_memcmp(path->start + path->length - 8, "/restart", 8) != 0)
+        || memcmp(path->start + path->length - 8, "/restart", 8) != 0)
     {
         goto not_found;
     }
@@ -1963,13 +2503,13 @@ nxt_controller_response(nxt_task_t *task, nxt_controller_request_t *req,
     nxt_conf_value_t        *value, *location;
     nxt_conf_json_pretty_t  pretty;
 
-    static nxt_str_t  success_str = nxt_string("success");
-    static nxt_str_t  error_str = nxt_string("error");
-    static nxt_str_t  detail_str = nxt_string("detail");
-    static nxt_str_t  location_str = nxt_string("location");
-    static nxt_str_t  offset_str = nxt_string("offset");
-    static nxt_str_t  line_str = nxt_string("line");
-    static nxt_str_t  column_str = nxt_string("column");
+    static const nxt_str_t  success_str = nxt_string("success");
+    static const nxt_str_t  error_str = nxt_string("error");
+    static const nxt_str_t  detail_str = nxt_string("detail");
+    static const nxt_str_t  location_str = nxt_string("location");
+    static const nxt_str_t  offset_str = nxt_string("offset");
+    static const nxt_str_t  line_str = nxt_string("line");
+    static const nxt_str_t  column_str = nxt_string("column");
 
     static nxt_time_string_t  date_cache = {
         (nxt_atomic_uint_t) -1,
@@ -2129,11 +2669,12 @@ static u_char *
 nxt_controller_date(u_char *buf, nxt_realtime_t *now, struct tm *tm,
     size_t size, const char *format)
 {
-    static const char  *week[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
-                                   "Sat" };
+    static const char * const  week[] = { "Sun", "Mon", "Tue", "Wed", "Thu",
+                                          "Fri", "Sat" };
 
-    static const char  *month[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
-                                    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
+    static const char * const  month[] = { "Jan", "Feb", "Mar", "Apr", "May",
+                                           "Jun", "Jul", "Aug", "Sep", "Oct",
+                                           "Nov", "Dec" };
 
     return nxt_sprintf(buf, buf + size, format,
                        week[tm->tm_wday], tm->tm_mday,
