@@ -14,6 +14,9 @@
 #if (NXT_TLS)
 #include <nxt_cert.h>
 #endif
+#if (NXT_HAVE_NJS)
+#include <nxt_script.h>
+#endif
 
 #include <sys/mount.h>
 
@@ -48,6 +51,8 @@ static void nxt_main_process_signal_handler(nxt_task_t *task, void *obj,
 static void nxt_main_process_cleanup(nxt_task_t *task, nxt_process_t *process);
 static void nxt_main_port_socket_handler(nxt_task_t *task,
     nxt_port_recv_msg_t *msg);
+static void nxt_main_port_socket_unlink_handler(nxt_task_t *task,
+    nxt_port_recv_msg_t *msg);
 static nxt_int_t nxt_main_listening_socket(nxt_sockaddr_t *sa,
     nxt_listening_socket_t *ls);
 static void nxt_main_port_modules_handler(nxt_task_t *task,
@@ -118,6 +123,18 @@ static nxt_conf_map_t  nxt_common_app_conf[] = {
         offsetof(nxt_common_app_conf_t, group),
     },
 
+    {
+        nxt_string("stdout"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, stdout_log),
+    },
+
+    {
+        nxt_string("stderr"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, stderr_log),
+    },
+
     {
         nxt_string("working_directory"),
         NXT_CONF_MAP_CSTRZ,
@@ -190,18 +207,6 @@ static nxt_conf_map_t  nxt_python_app_conf[] = {
         offsetof(nxt_common_app_conf_t, u.python.path),
     },
 
-    {
-        nxt_string("module"),
-        NXT_CONF_MAP_STR,
-        offsetof(nxt_common_app_conf_t, u.python.module),
-    },
-
-    {
-        nxt_string("callable"),
-        NXT_CONF_MAP_CSTRZ,
-        offsetof(nxt_common_app_conf_t, u.python.callable),
-    },
-
     {
         nxt_string("protocol"),
         NXT_CONF_MAP_STR,
@@ -318,6 +323,74 @@ static nxt_conf_map_t  nxt_java_app_conf[] = {
 };
 
 
+static nxt_conf_map_t  nxt_wasm_app_conf[] = {
+    {
+        nxt_string("module"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.module),
+    },
+    {
+        nxt_string("request_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.request_handler),
+    },
+    {
+        nxt_string("malloc_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.malloc_handler),
+    },
+    {
+        nxt_string("free_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.free_handler),
+    },
+    {
+        nxt_string("module_init_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.module_init_handler),
+    },
+    {
+        nxt_string("module_end_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.module_end_handler),
+    },
+    {
+        nxt_string("request_init_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.request_init_handler),
+    },
+    {
+        nxt_string("request_end_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.request_end_handler),
+    },
+    {
+        nxt_string("response_end_handler"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm.response_end_handler),
+    },
+    {
+        nxt_string("access"),
+        NXT_CONF_MAP_PTR,
+        offsetof(nxt_common_app_conf_t, u.wasm.access),
+    },
+};
+
+
+static nxt_conf_map_t  nxt_wasm_wc_app_conf[] = {
+    {
+        nxt_string("component"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_common_app_conf_t, u.wasm_wc.component),
+    },
+    {
+        nxt_string("access"),
+        NXT_CONF_MAP_PTR,
+        offsetof(nxt_common_app_conf_t, u.wasm_wc.access),
+    },
+};
+
+
 static nxt_conf_app_map_t  nxt_app_maps[] = {
     { nxt_nitems(nxt_external_app_conf),  nxt_external_app_conf },
     { nxt_nitems(nxt_python_app_conf),    nxt_python_app_conf },
@@ -325,6 +398,8 @@ static nxt_conf_app_map_t  nxt_app_maps[] = {
     { nxt_nitems(nxt_perl_app_conf),      nxt_perl_app_conf },
     { nxt_nitems(nxt_ruby_app_conf),      nxt_ruby_app_conf },
     { nxt_nitems(nxt_java_app_conf),      nxt_java_app_conf },
+    { nxt_nitems(nxt_wasm_app_conf),      nxt_wasm_app_conf },
+    { nxt_nitems(nxt_wasm_wc_app_conf),   nxt_wasm_wc_app_conf },
 };
 
 
@@ -382,25 +457,25 @@ nxt_main_start_process_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
     port = rt->port_by_type[NXT_PROCESS_ROUTER];
     if (nxt_slow_path(port == NULL)) {
         nxt_alert(task, "router port not found");
-        return;
+        goto close_fds;
     }
 
     if (nxt_slow_path(port->pid != nxt_recv_msg_cmsg_pid(msg))) {
         nxt_alert(task, "process %PI cannot start processes",
                   nxt_recv_msg_cmsg_pid(msg));
 
-        return;
+        goto close_fds;
     }
 
     process = nxt_process_new(rt);
     if (nxt_slow_path(process == NULL)) {
-        return;
+        goto close_fds;
     }
 
     process->mem_pool = nxt_mp_create(1024, 128, 256, 32);
     if (process->mem_pool == NULL) {
         nxt_process_use(task, process, -1);
-        return;
+        goto close_fds;
     }
 
     process->parent_port = rt->port_by_type[NXT_PROCESS_MAIN];
@@ -422,6 +497,9 @@ nxt_main_start_process_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
         goto failed;
     }
 
+    app_conf->shared_port_fd = msg->fd[0];
+    app_conf->shared_queue_fd = msg->fd[1];
+
     start = b->mem.pos;
 
     app_conf->name.start = start;
@@ -509,6 +587,17 @@ nxt_main_start_process_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
 
     ret = nxt_process_start(task, process);
     if (nxt_fast_path(ret == NXT_OK || ret == NXT_AGAIN)) {
+
+        /* Close shared port fds only in main process. */
+        if (ret == NXT_OK) {
+            nxt_fd_close(app_conf->shared_port_fd);
+            nxt_fd_close(app_conf->shared_queue_fd);
+        }
+
+        /* Avoid fds close in caller. */
+        msg->fd[0] = -1;
+        msg->fd[1] = -1;
+
         return;
     }
 
@@ -523,6 +612,14 @@ failed:
         nxt_port_socket_write(task, port, NXT_PORT_MSG_RPC_ERROR,
                               -1, msg->port_msg.stream, 0, NULL);
     }
+
+close_fds:
+
+    nxt_fd_close(msg->fd[0]);
+    msg->fd[0] = -1;
+
+    nxt_fd_close(msg->fd[1]);
+    msg->fd[1] = -1;
 }
 
 
@@ -546,7 +643,7 @@ nxt_main_process_created_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
     nxt_assert(process != NULL);
     nxt_assert(process->state == NXT_PROCESS_STATE_CREATING);
 
-#if (NXT_HAVE_CLONE && NXT_HAVE_CLONE_NEWUSER)
+#if (NXT_HAVE_LINUX_NS && NXT_HAVE_CLONE_NEWUSER)
     if (nxt_is_clone_flag_set(process->isolation.clone.flags, NEWUSER)) {
         if (nxt_slow_path(nxt_clone_credential_map(task, process->pid,
                                                    process->user_cred,
@@ -557,7 +654,7 @@ nxt_main_process_created_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
                                          -1, msg->port_msg.stream, 0, NULL);
             return;
         }
-     }
+    }
 
 #endif
 
@@ -577,11 +674,16 @@ static nxt_port_handlers_t  nxt_main_process_port_handlers = {
     .remove_pid       = nxt_port_remove_pid_handler,
     .start_process    = nxt_main_start_process_handler,
     .socket           = nxt_main_port_socket_handler,
+    .socket_unlink    = nxt_main_port_socket_unlink_handler,
     .modules          = nxt_main_port_modules_handler,
     .conf_store       = nxt_main_port_conf_store_handler,
 #if (NXT_TLS)
     .cert_get         = nxt_cert_store_get_handler,
     .cert_delete      = nxt_cert_store_delete_handler,
+#endif
+#if (NXT_HAVE_NJS)
+    .script_get       = nxt_script_store_get_handler,
+    .script_delete    = nxt_script_store_delete_handler,
 #endif
     .access_log       = nxt_main_port_access_log_handler,
     .rpc_ready        = nxt_port_rpc_handler,
@@ -769,7 +871,7 @@ nxt_main_process_sigusr1_handler(nxt_task_t *task, void *obj, void *data)
     nxt_array_t     *new_files;
     nxt_runtime_t   *rt;
 
-    nxt_log(task, NXT_LOG_NOTICE, "signal %d (%s) recevied, %s",
+    nxt_log(task, NXT_LOG_NOTICE, "signal %d (%s) received, %s",
             (int) (uintptr_t) obj, data, "log files rotation");
 
     rt = task->thread->runtime;
@@ -986,7 +1088,7 @@ nxt_main_process_sigchld_handler(nxt_task_t *task, void *obj, void *data)
 static void
 nxt_main_process_signal_handler(nxt_task_t *task, void *obj, void *data)
 {
-    nxt_trace(task, "signal signo:%d (%s) recevied, ignored",
+    nxt_trace(task, "signal signo:%d (%s) received, ignored",
               (int) (uintptr_t) obj, data);
 }
 
@@ -997,6 +1099,10 @@ nxt_main_process_cleanup(nxt_task_t *task, nxt_process_t *process)
     if (process->isolation.cleanup != NULL) {
         process->isolation.cleanup(task, process);
     }
+
+    if (process->isolation.cgroup_cleanup != NULL) {
+        process->isolation.cgroup_cleanup(task, process);
+    }
 }
 
 
@@ -1165,19 +1271,23 @@ nxt_main_listening_socket(nxt_sockaddr_t *sa, nxt_listening_socket_t *ls)
 
 #if (NXT_HAVE_UNIX_DOMAIN)
 
-    if (sa->u.sockaddr.sa_family == AF_UNIX) {
-        char     *filename;
-        mode_t   access;
+    if (sa->u.sockaddr.sa_family == AF_UNIX
+        && sa->u.sockaddr_un.sun_path[0] != '\0')
+    {
+        char          *filename;
+        nxt_thread_t  *thr;
 
         filename = sa->u.sockaddr_un.sun_path;
-        access = (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
 
-        if (chmod(filename, access) != 0) {
+        if (chmod(filename, 0666) != 0) {
             ls->end = nxt_sprintf(ls->start, ls->end,
                                   "chmod(\\\"%s\\\") failed %E",
                                   filename, nxt_errno);
             goto fail;
         }
+
+        thr = nxt_thread();
+        nxt_runtime_listen_socket_add(thr->runtime, sa);
     }
 
 #endif
@@ -1194,6 +1304,49 @@ fail:
 }
 
 
+static void
+nxt_main_port_socket_unlink_handler(nxt_task_t *task, nxt_port_recv_msg_t *msg)
+{
+#if (NXT_HAVE_UNIX_DOMAIN)
+    size_t               i;
+    nxt_buf_t            *b;
+    const char           *filename;
+    nxt_runtime_t        *rt;
+    nxt_sockaddr_t       *sa;
+    nxt_listen_socket_t  *ls;
+
+    b = msg->buf;
+    sa = (nxt_sockaddr_t *) b->mem.pos;
+
+    filename = sa->u.sockaddr_un.sun_path;
+    unlink(filename);
+
+    rt = task->thread->runtime;
+
+    for (i = 0; i < rt->listen_sockets->nelts; i++) {
+        const char  *name;
+
+        ls = (nxt_listen_socket_t *) rt->listen_sockets->elts + i;
+        sa = ls->sockaddr;
+
+        if (sa->u.sockaddr.sa_family != AF_UNIX
+            || sa->u.sockaddr_un.sun_path[0] == '\0')
+        {
+            continue;
+        }
+
+        name = sa->u.sockaddr_un.sun_path;
+        if (strcmp(name, filename) != 0) {
+            continue;
+        }
+
+        nxt_array_remove(rt->listen_sockets, ls);
+        break;
+    }
+#endif
+}
+
+
 static nxt_conf_map_t  nxt_app_lang_module_map[] = {
     {
         nxt_string("type"),
@@ -1201,6 +1354,12 @@ static nxt_conf_map_t  nxt_app_lang_module_map[] = {
         offsetof(nxt_app_lang_module_t, type),
     },
 
+    {
+        nxt_string("name"),
+        NXT_CONF_MAP_CSTRZ,
+        offsetof(nxt_app_lang_module_t, name),
+    },
+
     {
         nxt_string("version"),
         NXT_CONF_MAP_CSTRZ,
